package benno.globe;

import java.util.*;
import java.io.*;
import java.net.*;

/** 

@version $Revision: 1.11 $
@author Ben Clifford

*/


public class CurveCompiler 
{

	long totLen;
	long totSeg;

	public CurveSet parse(URL url) throws IOException
	{
		CurveSet curveset = new CurveSet();

		DataFileEnumeration en = new DataFileEnumeration(url,true);

		while(true)
		{

		Curve thisCurve = new Curve();

		int numPoints = en.nextInt();

		if(numPoints == 0) // we have reached the end.
		{
			return curveset;
		}

		boolean land = (en.nextInt() == 1);
		if(land)
		{
			thisCurve.isLand = true;
		} else {
			thisCurve.isLand = false;
		}

		long nowX = en.nextInt();
		long nowY = en.nextInt();
		long nowZ = en.nextInt();

		Position beginningPos;
		Position startPos;

		beginningPos = new Position(nowX,nowY,nowZ);	
		startPos = beginningPos;
		thisCurve.positions.addElement(startPos);	


		for(int i = 1; i<numPoints; i++)
		{
			long dX = en.nextInt();
			long dY = en.nextInt();
			long dZ = en.nextInt();
			nowX += dX;
			nowY += dY;
			nowZ += dZ;
			thisCurve.positions.addElement(new Position(nowX,nowY,nowZ));	

			// now compile statistics for this segment
			double length = Math.sqrt(dX*dX + dY*dY + dZ*dZ);
			totLen += length;
			totSeg ++;
		}
		thisCurve.positions.addElement(startPos=new Position(nowX,nowY,nowZ));	
		thisCurve.positions.addElement(beginningPos);	
	curveset.curves.addElement(thisCurve);
}
	}

	public CurveSet optimise(CurveSet inset, int t)
	{
// 300 works pretty well for 300x300 no-zoom
//		long threshold = 500;
		long threshold = t;
		CurveSet outset = new CurveSet();
		Enumeration curvesEnumeration = inset.curves.elements();
		while(curvesEnumeration.hasMoreElements())
		{
			Curve incurve = (Curve)curvesEnumeration.nextElement();
			Curve outcurve = new Curve();
			outcurve.isLand = incurve.isLand;

			Enumeration positionEnumeration = incurve.positions.elements();
			Position thisPos = (Position)positionEnumeration.nextElement();
			outcurve.positions.addElement(thisPos);
			while(positionEnumeration.hasMoreElements())
			{
				Position nextPos = (Position)positionEnumeration.nextElement();
				long delX, delY, delZ;
				delX = nextPos.X - thisPos.X;
				delY = nextPos.Y - thisPos.Y;
				delZ = nextPos.Z - thisPos.Z;
				long distance;
				distance = (long)Math.sqrt(delX*delX + delY*delY + delZ*delZ);
				if(distance>threshold || !positionEnumeration.hasMoreElements())
				{
					thisPos = nextPos;
					outcurve.positions.addElement(thisPos);
				}
			}
			outset.curves.addElement(outcurve);
		}
		return outset;
	}

}

