package benno.globe;

import java.util.Vector;
import java.util.zip.GZIPInputStream;
import java.io.*;
import java.net.*;

public class DataFileEnumeration
{

	public boolean debug = false;

	StreamTokenizer st;
	boolean ended = false;

	DataFileEnumeration(URL url) throws IOException
	{
		this(url,false);
	}

	DataFileEnumeration(URL url, boolean isCompressed) throws IOException
	{
			if(isCompressed)
				st = new StreamTokenizer(new BufferedReader(new InputStreamReader(new GZIPInputStream(url.openStream()))));
			else
				st = new StreamTokenizer(new BufferedReader(new InputStreamReader(url.openStream())));
			st.slashStarComments(true);
			st.slashSlashComments(true);
			st.commentChar('#');
		//	st.quoteChar('"');
	}

	String nextString() throws IOException
	{
			while(true) // loops until we return by finding a number
			{
				int i = st.nextToken();
				switch(i)
				{	
					case StreamTokenizer.TT_EOF: throw new IOException("Enumeration over"); 
					case '"':
					case StreamTokenizer.TT_WORD:
						if(debug) System.err.println("Correctly parsed "+st.sval);
						return st.sval;
					case StreamTokenizer.TT_NUMBER:
						System.err.println("Found a number "+st.nval+"; ignoring."); break;
					default: 
						if(i>32)
						{
							if(debug) System.out.println(""+(char)i);
						}
						else
							System.err.println("DEFAULT hit when looking for an string: "+i);
						break; // ignore
				}
			}

	}

	int nextInt() throws IOException
	{
			while(true) // loops until we return by finding a number
			{
				int i = st.nextToken();
				switch(i)
				{	
					case StreamTokenizer.TT_EOF: throw new IOException("Enumeration over"); 
					case '"':
					case StreamTokenizer.TT_WORD:
						System.err.println("Found a word "+st.sval+"; ignoring."); break;
					case StreamTokenizer.TT_NUMBER:
						if(debug) System.err.println("Correctly parsed #"+st.nval);
						return (int)st.nval;
					default: 
						if(i>32)
						{
							if(debug) System.out.println(""+(char)i);
						}
						else
							System.err.println("DEFAULT hit when looking for an int: "+i);
						break; // ignore
				}
			}

	}

	double nextDouble() throws IOException
	{
			while(true) // loops until we return by finding a number
			{
				int i = st.nextToken();
				switch(i)
				{	
					case StreamTokenizer.TT_EOF: throw new IOException("Enumeration over"); 
					case '"':
					case StreamTokenizer.TT_WORD:
						System.err.println("Found a word "+st.sval+"; ignoring."); break;
					case StreamTokenizer.TT_NUMBER:
						if(debug) System.err.println("Correctly parsed #"+st.nval);
						return (double)st.nval;
					case '@': // treat the .xx part as /60 not /100
						double tempd = nextDouble();
						int ipart = (int)tempd;
						double fpart = tempd-ipart;
						fpart/=100;
						fpart*=60;
						return ipart + fpart;
					default: 
						if(i>32)
						{
							if(debug) System.out.println(""+(char)i);
						}
						else
							System.err.println("DEFAULT hit when looking for a double: "+i);
						break; // ignore
				}
			}

	}

}
