//$Id: Globe2.java,v 1.21 2000/09/26 22:31:56 benc Exp $
package benno.globe;

import java.net.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import benno.sort.*;

/** Displays a Globe with labels.
this version does not attempt to fill in the outlines as the curve 
filling code doesn't work

@version $Revision: 1.21 $

 */

public class Globe2 extends java.applet.Applet implements Runnable
{
	private long rotateDelay = 10000; //ms
	private int rotateAngle = 0; // degs/frame
	private int rotateAngleLat = 10; // degs/frame  -- doesn't work properly
	private boolean randomizeEachFrame = false; 
	private boolean randomizeFirstFrame = false;
	private boolean filled = false;
	private int optimiseLength = 300; // in world coords
        private String target = null;

        private boolean suspended = false;

	private String statusDisplay = "Please wait...";

	private Graphics graphicsContext = null;

	int paints=0;

	Thread renderThread;
/** doubleBuffer is the buffer that we draw into */
//	Image doubleBuffer = null;

/** validBuffer is the buffer that is displayed. Once we have a valid
image, it will always contain a valid image - it will not be updated 
except to replace it with a new valid image. */
	Image validBuffer = null;
	Image shown = null;

	Thread triggerThread;
	VectorRandomiser vectorRandomiser = new VectorRandomiser();
	Color globeColor = new Color(0,0,40);
	Color outlineSeaColor = new Color(40,80,128);
	Color outlineLandColor = new Color(40,128,40);
	Color markerColor = new Color(232,200,200);
	LabelPlane displayedLabelPlane;
	LabelPlane labelPlane;

	int dimension;
        int scaledim;

	CurveSet curves = null;
	MarkerSet markers = null;
        Marker northpole = null;
	Font markerFont = new Font("SansSerif",Font.PLAIN,10);
	FontMetrics markerFontMetrics = getFontMetrics(markerFont);
	boolean[][] pixels;

	LongLat centeredOn = new LongLat(0,0);

	public void init()
	{
		showStatus("Initialising...");

		try
		{
			rotateDelay = Long.parseLong(getParameter("rotateDelay"));
			if(rotateDelay < 1) rotateDelay = 1;
		} catch(Exception e) {;}

		try
		{
			rotateAngle = Integer.parseInt(getParameter("rotateAngle"));
		} catch(Exception e) {;}

		try
		{
			optimiseLength = Integer.parseInt(getParameter("optimiseLength"));
		} catch(Exception e) {;}

		try
		{
			if(null!=getParameter("randomizeEachFrame"))
				randomizeEachFrame = Boolean.valueOf(getParameter("randomizeEachFrame")).booleanValue();
		} catch(Exception e) {;}

		try
		{
			if(null!=getParameter("randomizeFirstFrame"))
				randomizeFirstFrame = Boolean.valueOf(getParameter("randomizeFirstFrame")).booleanValue();
		} catch(Exception e) {;}

		try
		{
			if(null!=getParameter("target"))
				target = getParameter("target");
		} catch(Exception e) {;}

		try
		{
			if(null!=getParameter("filled"))
				filled = Boolean.valueOf(getParameter("filled")).booleanValue();
		} catch(Exception e) {;}

		addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent e)
			{
				mouseClick(e); // pass to outer class
			}
		});

		renderThread = new Thread(this);
		renderThread.start();
	}

	public void thread_init()
	{
		Dimension s = getSize();
		// Dimension s = size();
		dimension = (s.height < s.width) ? s.height : s.width;
                scaledim = (int)dimension;
		CurveCompiler ccomp = new CurveCompiler();
		MarkerProvider mcomp = new MarkerCompiler();
                northpole = new Marker(null, new LongLat(0,90), null, null);
		try {
			showStatus("Loading map");
			URL dataURL = new URL(getDocumentBase(),getParameter("mapdata"));
			curves = ccomp.parse(dataURL);
			showStatus("Loading markers");
			dataURL = new URL(getDocumentBase(),getParameter("markerdata"));
			markers = mcomp.get(dataURL);
			if(randomizeFirstFrame)
				vectorRandomiser.randomise(markers.markers);
			showStatus("Map & Markers loaded");
		} catch(Exception e)
		{
			System.err.println("Got an exception: "+e);
			showStatus("Exception: "+e);
			return;
		}
		showStatus("Optimising map");
		curves = ccomp.optimise(curves, optimiseLength);
		showStatus("Map optimised");
	}

	public void run()
	{
		thread_init();
                thread_paint();
                /**
		triggerThread = new Thread() 
		{ 
			long thrashDelay = rotateDelay;
			public void run() 
			{ 
				while(true)
				{
						thread_paint();
						centeredOn = new LongLat(centeredOn.longitude+rotateAngle, centeredOn.latitude+rotateAngleLat);
						thrashDelay = rotateDelay;
/*
					if(shown==validBuffer) 
					{
						thread_paint();
						centeredOn = new LongLat(centeredOn.longitude+rotateAngle, centeredOn.latitude+rotateAngleLat);
						thrashDelay = rotateDelay;
					}
					else
					{
						thrashDelay = 4 * thrashDelay;
						System.out.println("ThrashDelay doubled to: "+thrashDelay+"ms");
						//System.out.println("shown="+shown+", validBuffer="+validBuffer);
						//System.out.println("---");
						repaint();
					}
*/
/**
					try {
						Thread.sleep(thrashDelay);
					} catch(InterruptedException e) { ; } // do nothing
				}
			}
		};
		triggerThread.setDaemon(true);
		triggerThread.start();
                */
		showStatus("");
	}
        public LongLat FindNorthangle(Marker northpole, LongLat pos)
        {
                double northangle;
                LongLat zeroangle = new LongLat(0,0);
                Marker testpole = rotateMarkerTo(northpole,pos,zeroangle);
                Position testpolePos = cartesian(testpole.pos);
                if (testpolePos.X == 0.0 && testpolePos.Y == 0.0) 
                {
                   northangle = 0.0;
                }
                else 
                {
                   northangle = Math.atan2((double)testpolePos.X,(double)testpolePos.Y);
                }
                return new LongLat(northangle / (2.0 * Math.PI) * 360.0, 0.0);
        }

	public MarkerSet rotateTo(MarkerSet s, LongLat pos, LongLat northangle)
	{
		MarkerSet ms = new MarkerSet();
		Enumeration e = s.markers.elements();
		while(e.hasMoreElements())
		{
			Marker newMarker = rotateMarkerTo((Marker)e.nextElement(),pos,northangle);
			ms.markers.addElement(newMarker);
		}
		return ms;
	}

	public CurveSet rotateTo(CurveSet s, LongLat pos, LongLat northangle)
	{
		CurveSet cs = new CurveSet();
		Enumeration e = s.curves.elements();
		while(e.hasMoreElements())
		{
			Curve newCurve = rotateCurveTo((Curve)e.nextElement(),pos,northangle);
			cs.curves.addElement(newCurve);
		}
		return cs;
	}

        public Position cartesian(LongLat p)
        {
                double R = 30000.0;
                double mX = R * p.lat_cos * p.long_sin;
                double mY = R * p.lat_sin;
                double mZ = R * p.lat_cos * p.long_cos; 
		return new Position((long)mX,(long)mY,(long)mZ);
        }

        public LongLat polar(Position p)
        {
                double R = 30000.0;
                double dlat = Math.asin( (double)p.Y/R );
                double dlongi = Math.asin( (double)p.X/R / Math.cos(dlat) );
                dlat = dlat / (2.0 * Math.PI) * 360.0;
                dlongi = dlongi / (2.0 * Math.PI) * 360.0 ;
                if (p.Z < 0)
                {
                   if (dlongi > 0)
                   {
                     dlongi = 180.0 - dlongi;
                   }
                   else
                   {
                     dlongi = -180.0 - dlongi;
                   }
                }
		return new LongLat(dlongi,dlat);
        }

	public Marker rotateMarkerTo(Marker m, LongLat pos, LongLat northangle)
	{
		//double newlong = m.pos.longitude + p.longitude;

		//double newlat = m.pos.latitude + p.latitude;
		// double newlat = m.pos.latitude;

                Position newp = rotatePosition(cartesian(m.pos),pos,northangle);
		LongLat newpos = polar(newp);

                //showStatus(m.desc+": "+p.X+":"+p.Y+":"+p.Z+" polar: "+newpos.longitude+":"+newpos.latitude);

		return new Marker(m.desc, newpos, m.url, m.color);
	}

	public Curve rotateCurveTo(Curve c, LongLat pos, LongLat northangle)
	{
		Curve newc = new Curve();
		newc.isLand = c.isLand;
		Enumeration e = c.positions.elements();

		while(e.hasMoreElements())
		{
			Position p = (Position)e.nextElement();
			newc.positions.addElement(rotatePosition(p,pos,northangle));
		}
		return newc;
	}

	Position rotatePosition(Position p, LongLat pos, LongLat northangle)
	{

		double nX = ((double)p.X)*pos.long_cos + ((double)p.Z)*pos.long_sin;
		double nY = p.Y;
		double nZ = ((double)p.X)*pos.long_sin*-1.0 + ((double)p.Z)*pos.long_cos;


		double mX = nX   ;
		double mY = nY * pos.lat_cos + nZ*pos.lat_sin ;
		double mZ = nY * pos.lat_sin * -1.0 + nZ*pos.lat_cos ;
	
		double lX = mX * northangle.long_cos - mY * northangle.long_sin;
		double lY = mX * northangle.long_sin + mY * northangle.long_cos;
		double lZ = mZ;

		// return new Position((long)nX,(long)nY,(long)nZ);
		return new Position((long)lX,(long)lY,(long)lZ);
	}

/** special override for double-buffering. just calls paint, without
    clearing the area. */
	public void update(Graphics g)
	{
		graphicsContext = g;
		paint(g);
	}

	public void paint(Graphics g)
	{
		if(validBuffer==null)	
		{
			g.setColor(Color.white);
			g.fillRect(0,0,getSize().width, getSize().height);
			g.setColor(Color.black);
			g.drawString(statusDisplay,50,50);
		}
		else
		{
			g.drawImage(validBuffer,0,0,null);
			shown = validBuffer;
			synchronized(this) {
				paints--;
			}
		}
	}

	public void thread_paint()
	{
		if(paints>0) 
		{
				System.out.println("Outstanding paints:"+paints);
				repaint();
				return;
		}
		showStatus("Painting");
		Image doubleBuffer = createImage(dimension,dimension);
		Graphics g = doubleBuffer.getGraphics();
		g.setColor(Color.white);
		g.fillRect(0,0,dimension,dimension);
		//g.setColor(Color.black);
		//g.drawRect(0,0,dimension-1,dimension-1);
		labelPlane = new LabelPlane();	

                LongLat northangle = FindNorthangle(northpole,centeredOn);
                northpole = rotateMarkerTo(northpole,centeredOn,northangle);
		curves = rotateTo(curves,centeredOn,northangle);
		CurveSet processed = curves;
		//MarkerSet markerProcessed = rotateTo(markers,centeredOn,northangle);
		markers = rotateTo(markers,centeredOn,northangle);
                MarkerSet markerProcessed = markers;

		drawGlobeOutline(g);

		Enumeration e = processed.curves.elements();
		int i=0;
		while(e.hasMoreElements())
		{
			Curve c = (Curve)e.nextElement();
		
			if(filled)
				drawFilledCurve(c,g);
			drawOutlineCurve(c,g);
		}
		if(randomizeEachFrame)
			vectorRandomiser.randomise(markers.markers);
		e=markerProcessed.markers.elements();
		while(e.hasMoreElements())
		{
			Marker m = (Marker)e.nextElement();
			drawMarker(m,g);
		}
		showStatus("finished painting");
		validBuffer = doubleBuffer;
		doubleBuffer = null;
		displayedLabelPlane = labelPlane;
		synchronized(this)
		{
			paints++;
		}
		repaint();
	}

	private void drawMarker(Marker m, Graphics g)
	{

		if(m.pos.longitude > 90 || m.pos.longitude<-90) // not visible
			return;

		// if we are here, then the marker is visible somewhere

		double mx = 30000.0 * m.pos.lat_cos * m.pos.long_sin;
		double my = 30000.0 * m.pos.lat_sin;

		int sx = (int)mx*scaledim/60000+dimension/2;
		int sy = (int)my*scaledim/-60000+dimension/2;

		// determine bounding rectangle for marker
		
		int markTop, markRight, markLeft, markBottom;

		markBottom = sy + markerFontMetrics.getDescent();
		markLeft = sx - markerFontMetrics.stringWidth(m.desc)/2;
		markRight = sx + markerFontMetrics.stringWidth(m.desc)/2;
		markTop = sy - markerFontMetrics.getAscent();

		Rectangle textbound = new Rectangle(markLeft,markTop,markRight-markLeft+2, markBottom-markTop);

		if(!labelPlane.checkAllocated(textbound))
		{
			g.setColor(Color.black);
			g.fillRect(textbound.x, textbound.y, textbound.width, textbound.height);
			g.setColor(Color.blue);
			g.drawRect(textbound.x, textbound.y, textbound.width, textbound.height);
			if(null!=m.color)
				g.setColor(m.color);
			else
				g.setColor(markerColor);
			g.setFont(markerFont);
			g.drawString(m.desc, markLeft+2, sy);
			g.setColor(Color.red);
			//g.setColor(m.color);
			g.drawOval(sx-1,sy-1,3,3);
			labelPlane.allocate(textbound, m.url );
		}
	}

	/** draws the globe as sea ready for things to be drawn on top */
	private void drawGlobeOutline(Graphics g)
	{
		g.setColor(globeColor);
		for(int y=0; y<dimension; y++)
		{
			long worldy = (y - (dimension/2) ) * -60000/scaledim;
                        if (Math.abs(worldy) <= 30000)
                        {
			   double xbound = Math.sqrt(Math.pow(30000,2)-Math.pow(worldy,2));
			   g.drawLine(
			   	   (int)xbound*scaledim/60000+dimension/2,
				   y,
				   (int)xbound*scaledim/-60000+dimension/2,
				   y);
                        }

		}
	}

	/** draws the specified Curve onto the specified Graphics object */
	private void drawOutlineCurve(Curve c, Graphics g)
	{
		Enumeration e = c.positions.elements();
		if(c.isLand) g.setColor(outlineLandColor); else g.setColor(outlineSeaColor);
		Position a = (Position)e.nextElement();
		while(e.hasMoreElements())
		{
			
			Position b = (Position)e.nextElement();
			if(a.Z>0 && b.Z>0) g.drawLine(
				(int) a.X*scaledim/60000+dimension/2,
				(int) a.Y*scaledim/-60000+dimension/2,
				(int) b.X*scaledim/60000+dimension/2,
				(int) b.Y*scaledim/-60000+dimension/2
			);
			a=b;
		}
	}
/** fills the curve, used red for land and cyan for sea */
	private void drawFilledCurve(Curve c, Graphics g)
	{
		if(c.isLand) g.setColor(Color.red); else g.setColor(Color.cyan);

		for(int screeny=0; screeny<dimension; screeny++)
		{
			double[] crossings = new double[30]; // can have a maximum of 30 crossings
			boolean[] crossbool = new boolean[30]; // can have a maximum of 30 crossings
			// crossbool is true if the crossing is from inside to outside
			int cridx = 0;

/** 1. changing this to double rather than long calculation did not
       appear to have any visible effect. */

			// long worldy = (screeny - (dimension/2) ) * -60000/scaledim;
			long worldy = (long)( 
			  ( ((double)screeny) - ( ((double)dimension) /2.0) ) 
			    * -60000.0/((double)scaledim)
			  ) ;

			// so now we can filter down to the few curves that intersect this
			// horizontal line
			Vector v = new Vector();
			Enumeration e = c.positions.elements();
			Position n = (Position)e.nextElement();
			Position p;
			while(e.hasMoreElements())
			{
				p=n;
				n=(Position)e.nextElement();
				if(n.Z < 0 || p.Z < 0) continue; // ignore because goes behind?
				// so at this stage, we know that both points are in the foreground
				if(n.Y > worldy && p.Y > worldy) continue;
				if(n.Y < worldy && p.Y < worldy) continue;

				if(n.Y == p.Y) continue; // if the line is horizontal

				// so at this point, we know that the line between the points
				// intersects the worldy line
				// so we can compute an intersection




		//? is this valid - I have not checked for a long time...
// 2. should check this
	
				/* double ix = ((double)(n.X - p.X))
				                      /
                    ((double)(n.Y - p.Y)) * ((double)(worldy - n.Y)) 


                   - (double)n.X;
*/

				double ix = ((double)(n.X - p.X)) * ((double)(n.Y - worldy))
				                      /
                    ((double)(n.Y - p.Y)) 

                   - (double)n.X;


				crossings[cridx]=ix;
				crossbool[cridx]=(p.Y>n.Y);

				cridx++;
			}

			if(cridx!=0) 
			{
				// now sort by xpos
				sort(crossings,crossbool,cridx); 
			
				// now for safety, check that crossings is ordered
				{
					double sofar = crossings[0];
					for(int i=1;i<cridx;i++)
					if(crossings[i]<sofar)
						throw new RuntimeException("Sort returned unsorted data");
					else
						sofar = crossings[i];
				}

				double xbound = Math.sqrt(30000.0*30000.0-(double)(worldy*worldy));
	
				// now render
				// int starting=0;
				int starting=(int)xbound*scaledim/60000+dimension/2;
				int inside; // more +ve every time we enter
				// inside = crossbool[0] ? 1 : 0;
				inside = 0;
				for(int i=0;i<cridx;i++)
				{
					if(crossbool[i] && inside>0)
					{
						g.drawLine(starting,screeny,(int)crossings[i]*scaledim/-60000+dimension/2,screeny);
						inside--;
					} else if(!crossbool[i] && inside<=0)  
					{
						starting=(int)crossings[i]*scaledim/-60000+dimension/2;
						inside++;
					} else if(crossbool[i] && inside<=0)
					{
						inside--;
					} else if(!crossbool[i] && inside>0)
					{
						starting=(int)crossings[i]*scaledim/-60000+dimension/2;
						inside++;
					} else throw new RuntimeException("Uncovered case in big if ??");
				}
				if(inside>0)
				{
						g.drawLine(starting,screeny,(int)xbound*scaledim/-60000+dimension/2,screeny);
				}

			}
		}
	}

	private void sort(double[] data, boolean[] booldata, int m)
	{
		for(int i=0;i<m;i++)
		{
			int minpos = i;
			for(int j=i+1; j<m; j++)
			{
				if(data[j]<data[minpos]) minpos = j;
			}
			// so now maxpos indicates the largest element of data
			double temp = data[i];
			data[i]=data[minpos];
			data[minpos] = temp;
			boolean bt = booldata[i];
			booldata[i]=booldata[minpos];
			booldata[minpos] = bt;
		}
	}

	Color getColorPref(String param)
	{
		int color = Integer.parseInt(getParameter(param),16);
		return new Color(color);
	}

	public void start()
	{
		if(renderThread!=null && renderThread.isAlive()) renderThread.resume();
		if(triggerThread!=null && triggerThread.isAlive()) triggerThread.resume();
	}

	public void stop()
	{
		if(renderThread!=null && renderThread.isAlive()) renderThread.suspend();
		if(triggerThread!=null && triggerThread.isAlive()) triggerThread.suspend();
	}

	public void destroy()
	{
		if(renderThread!=null) 
		{
			renderThread.stop();
			renderThread=null; // so we can gc
		}
		if(triggerThread!=null) 
		{
			triggerThread.stop();
			triggerThread=null; // so we can gc
		}
	}

	public void showStatus(String s)
	{
		statusDisplay = s;
		super.showStatus(statusDisplay);
		if(validBuffer==null) repaint();

	}

	public void mouseClick(MouseEvent e)
	{
           if ((e.getModifiers() & InputEvent.BUTTON3_MASK) == InputEvent.BUTTON3_MASK)                 
           {
		System.out.println("Mouse right-clicked... looking for point "+e.getPoint());
		URL annotation = (URL)labelPlane.getAnnotationFor(e.getPoint());
		if(annotation!=null) 
		{
			System.out.println("Going to "+annotation);
                        if(target!=null)
                        {
			     getAppletContext().showDocument(annotation, target);
                        }
                        else {
			     getAppletContext().showDocument(annotation);
                        }
		}
            }
            else 
            {
                if (e.isShiftDown())
                {
                    scaledim = (int)(scaledim * 1.2);
                    thread_paint();
                }
                else
                {
                    if (e.isControlDown())
                    {
                       scaledim = (int)(scaledim / 1.2);
                       thread_paint();
                    }
                    else {
                       Point clickPoint = anglesofPoint(e.getPoint());
        	       System.out.println("Mouse clicked once... looking for point "+e.getPoint());
                       if (clickPoint != null)
                       {
                          centeredOn = new LongLat(-1 * clickPoint.x,
                                                -1 * clickPoint.y);
                          thread_paint();                
                          centeredOn = new LongLat(0,0);
                       }
                    }
                }
            }
	}

	private Point anglesofPoint(Point p)
	{
                int x = p.x;
                int y = p.y;
		float worldy = (y - (dimension/2) ) * -60000/scaledim;
		float worldx = (x - (dimension/2) ) * 60000/scaledim;
                if (Math.sqrt(Math.pow(worldx,2)+Math.pow(worldy,2)) >= 30000)
                { 
                   return null;
                }
                double dlat = Math.asin( worldy/30000 );
                double dlongi = Math.asin( worldx/30000 / Math.cos(dlat) );
                dlat = dlat / (2.0 * Math.PI) * 360.0;
                dlongi = dlongi / (2.0 * Math.PI) * 360.0 ;
                Float flat = new Float(dlat); 
                Float flongi = new Float(dlongi);
                int lat = flat.intValue();
                int longi = flongi.intValue();
                //showStatus("x, y of Click: "+x+":"+y);
                //showStatus("worldx, worldy of Click: "+worldx+":"+worldy);
                //showStatus("worldx, worldy of Click: "+worldx+":"+worldy+"; lat, longi of Click: "+lat+":"+longi);
                return new Point(longi,lat);
                
                
	}

}

