package benno.globe;

/** Represents a longitude/lattitude pair. Instances of this class
also store precomputed values for sin and cos of longitude and latitude */

public class LongLat
{
	public final double longitude;
	public final double latitude;

	public final double long_sin;
	public final double long_cos;
	public final double lat_sin;
	public final double lat_cos;

	public LongLat(double lo, double la)
	{
		while(la > 90)
		{
			// thats over the top - so we need to rotate round 180 degrees
			lo+=180; // rely on the next section to normalise this
			la=180-la; // flip this back
		}

		while(la < -90)
		{
			lo+=180; // rely on the next section to normalise this
			la=-180-la; // flip this back
		}

		while(lo < -180) lo += 360;
		while(lo > 180) lo -= 360;


		longitude = lo;
		latitude = la;

		double long_rad = lo / 360.0 * 2.0 * Math.PI;
		double lat_rad = la / 360.0 * 2.0 * Math.PI;

		long_sin = Math.sin(long_rad);
		long_cos = Math.cos(long_rad);
		lat_sin = Math.sin(lat_rad);
		lat_cos = Math.cos(lat_rad);
	}


}
